#!/usr/bin/env python
"""
Functional Test Suites for some functionalities in B-tree project.

Assumes that "make load" has already been called

Jiayu Liu - jiayul@andrew.cmu.edu
Feb 4th 2015
"""
import re
import subprocess
from subprocess import PIPE

BIN_PATH = "./main"

def run_binary(cmd):
    p = subprocess.Popen([BIN_PATH], stdin=PIPE, stdout=PIPE, stderr=PIPE)
    stdout, stderr = p.communicate(input=cmd)
    if p.poll():
        print >> stderr, "process did not exit properly"
    if stderr:
        print >> stderr, "stderr", stderr
    return stdout
    
def scan_all_words():
    all_words = run_binary("""
    C
    x
    """)
    return sorted(re.findall(r'^(\w+):\d+', all_words, re.MULTILINE))
    
def uni_gap(lst, size=10):
    return xrange(0, len(lst), len(lst) / (size - 1))
    
def test_search(all_words):
    """
    uniformly sample 10 words to test
    """
    for i in uni_gap(all_words):
        result = run_binary(cmd="""
        s {0}
        x
        """.format(all_words[i]))
        yield "Found the key!" in result
        
def test_successors(all_words):
    """
    uniformly sample 10 words to test
    """
    for i in uni_gap(all_words):
        for k in [-1, 0, 1, 3, 5, 10, 42, 100]:
            result = run_binary(cmd="""
            > {0} {1}
            x
            """.format(all_words[i], k))
            preds = set(re.findall(r'^(\w+)\s*$', result, re.MULTILINE))
            expected = set(all_words[i+1 : i+k+1])
            if preds != expected:
                print "user result", preds
                print "expected", expected
            yield preds == expected
        
def test_predecessors(all_words):
    """
    uniformly sample 10 words to test
    """
    for i in uni_gap(all_words):
        for k in [-1, 0, 1, 3, 5, 10, 42, 100]:
            result = run_binary(cmd="""
            < {0} {1}
            x
            """.format(all_words[i], k))
            preds = set(re.findall(r'^(\w+)\s*$', result, re.MULTILINE))
            expected = set(all_words[i-k : i])
            if preds != expected:
                print "user result", preds
                print "expected", expected
            yield preds == expected

def to_score(lst):
    return "score: {0}/{1}".format(sum(filter(lambda x: x is True, lst)), len(lst))

def main():
    all_words = scan_all_words()
    print "the b-tree contains", len(all_words), "words"
    
    print "search on some of then should always return true:"
    search_result = list(test_search(all_words))
    print to_score(search_result)
    
    print "search on some predecessors:"
    predecessors = list(test_predecessors(all_words))
    print to_score(predecessors)
    
    print "search on some successors:"
    successors = list(test_successors(all_words))
    print to_score(successors)
    
    
if __name__ == '__main__':
    main()